/* powerpc64le-linux.elf-entry.S -- Linux program entry point & decompressor (Elf binary)
*
*  This file is part of the UPX executable compressor.
*
*  Copyright (C) 1996-2017 Markus Franz Xaver Johannes Oberhumer
*  Copyright (C) 1996-2017 Laszlo Molnar
*  Copyright (C) 2000-2017 John F. Reiser
*  All Rights Reserved.
*
*  UPX and the UCL library are free software; you can redistribute them
*  and/or modify them under the terms of the GNU General Public License as
*  published by the Free Software Foundation; either version 2 of
*  the License, or (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; see the file COPYING.
*  If not, write to the Free Software Foundation, Inc.,
*  59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*
*  Markus F.X.J. Oberhumer              Laszlo Molnar
*  <markus@oberhumer.com>               <ezerotven+github@gmail.com>
*
*  John F. Reiser
*  <jreiser@users.sourceforge.net>
*/

#ifndef BIG_ENDIAN
#define BIG_ENDIAN 0
#endif

#include "arch/powerpc/64le/macros.S"
#include "arch/powerpc/64le/ppc_regs.h"

sz_b_info= 12
  sz_unc= 0
  sz_cpr= 4
  b_method= 8

PROT_READ=  1
PROT_WRITE= 2
PROT_EXEC=  4

MAP_PRIVATE= 2
MAP_FIXED=     0x10
MAP_ANONYMOUS= 0x20

PAGE_SHIFT= 16  // 64KiB PAGE_SIZE
PAGE_SIZE = -(~0<<PAGE_SHIFT)

/* /usr/include/asm-ppc/unistd.h */
__NR_write = 4
__NR_exit  = 1
__NR_mmap  = 90

  section ELFMAINX
_start: .globl _start
        call main  // must be exactly 1 instruction; link_register= &decompress

/* Returns 0 on success; non-zero on failure. */
decompress:  // (uchar const *src, size_t lsrc, uchar *dst, size_t &ldst, uint method)

  section NRV_HEAD
SZ_DLINE=128  # size of data cache line in Apple G5

/* PowerPC has no 'cmplis': compare logical [unsigned] immediate shifted [by 16] */
#define  hibit r0  /* holds 0x80000000 during decompress */

#define src  a0
#define lsrc a1
#define dst  a2
#define ldst a3  /* Out: actually a reference: &len_dst */
#define meth a4

#define off  a4
#define len  a5
#define bits a6
#define disp a7

#ifndef BIG_ENDIAN
#define BIG_ENDIAN 0
#endif

  section NRV2E
#include "arch/powerpc/64le/nrv2e_d.S"

  section NRV2D
#include "arch/powerpc/64le/nrv2d_d.S"

  section NRV2B
#include "arch/powerpc/64le/nrv2b_d.S"

// lzma decoder is pre-assembled and endian-dependent
#if BIG_ENDIAN
#include "arch/powerpc/64/lzma_d.S"
#else
#include "arch/powerpc/64le/lzma_d.S"
#endif

  section NRV_TAIL
eof_nrv:
#define dst0 a4
#define tmp a1
        ld dst0,0(ldst)  // original dst
        mtlr t3  // return address
        subf a0,lsrc,src
        subf tmp,dst0,dst  // -1+ dst length
        addi a0,a0,1  // return 0: good; else: bad  [+1: correct for lbzu]
        addi tmp,tmp,1  // dst length
        std  tmp,0(ldst)
#undef tmp

// CACHELINE=32 is the observed minimum line size of any cache.
// Some caches may have larger lines, but it is cumbersome to lookup
// {AT_DCACHEBSIZE, AT_ICACHEBSIZE, AT_UCACHEBSIZE: /usr/include/elf.h},
// then save the correct size in a variable {where to put it?}, or to modify
// the two instructions here.  If a cache has larger lines, then we expect
// that the second dcbst (or icbi) on a the same line will be fast.
// If not, then too bad.

  section CFLUSH  // In: a2=dst= &highest stored byte; a4=dst0= &lowest stored byte
CACHELINE=32
        ori dst0,dst0,-1+ CACHELINE  // highest addr on cache line
cfl_nrv:
        dcbst  0,dst0  // initiate store (modified) cacheline to memory
        cmpld cr0,dst0,dst  // did we cover the highest-addressed byte?
        icbi   0,dst0  // discard instructions from cacheline
        addi     dst0,dst0,CACHELINE  // highest addr on next line
        blt  cr0,cfl_nrv  // not done yet
#undef dst0
        sync   // wait for all memory operations to finish
        isync  // discard prefetched instructions (if any)
cfl_ret:
        ret

  section ELFMAINY
msg_SELinux:
        call L72
L70:
        .asciz "PROT_EXEC|PROT_WRITE failed.\n"
L71:
        // IDENTSTR goes here

#if BIG_ENDIAN  //{
  section ELFMAINZe
        .balign 8
        .globl entry_descr
entry_descr:  // Elf64_Ehdr.e_entry points here (big endian only)
        .quad _start,0,0
#endif  //}

  section ELFMAINZ
        .balign 4
L72:
        li   a2,L71 - L70  // length
        mflr a1    // message text
        li   a0,2  // fd stderr
        li 0,__NR_write; sc
die:
        li a0,127
        li 0,__NR_exit; sc

        /* Decompress the rest of this loader, and jump to it. */
unfold:
        mflr r30  // -4+ &{ b_info={sz_unc, sz_cpr, {4 char}}, folded_loader...}
        lwz r26,0(r30)  // O_BINFO
sz_obinfo= 4
             lwz  a0,sz_obinfo + sz_cpr(r30)  // sz_cpr(fold) (< 1 Page)
        li a5,0  // off_t
             lis  a1,1<<(PAGE_SHIFT - 16)  // PAGE_SIZE 64KiB
        li a4,-1  // fd; cater to *BSD for MAP_ANON
             add  a0,a0,r30  // beyond compressed fold
        li a3,MAP_PRIVATE | MAP_FIXED | MAP_ANONYMOUS
             addi a0,a0,sz_obinfo + sz_b_info-1
        li a2,PROT_READ | PROT_WRITE | PROT_EXEC
             add  a0,a0,a1  // across next page boundary
        li 0,__NR_mmap
             rldicr a0,a0,0,64 - PAGE_SHIFT  // next page boundary after fold
        sc; bso- msg_SELinux  // Branch if SummaryOverflow (failure)

        mtctr r31
        lwz r0,sz_obinfo + sz_unc(r30)        // size uncompressed
        lbz meth,sz_obinfo + b_method(r30)
        la ldst,SZ_FRAME+31*8(sp)  // &slot on stack
        stw  r0,SZ_FRAME+31*8(sp)  // lzma uses for EOF; BEWARE: 32 bits only
        mr dst,a0
        mtlr   a0  // &continuation
        lwz lsrc,sz_obinfo + sz_cpr(r30)        // size of compressed
        addi src,r30,sz_obinfo + sz_b_info
        stdu sp,-SZ_FRAME(sp)  // ABI calling convention (for lzma compiled by gcc)
        bctr  // goto decomrpess; return to link register (mmap'ed page)

main:
////  teq r0,r0  // debugging
        stdu r1,-(SZ_FRAME + 32*8)(sp)  // allocate space (keeping 0 mod 16), save r1
        std   2,SZ_FRAME+ 2*8-8(sp)
        std   3,SZ_FRAME+ 3*8-8(sp)
        std   4,SZ_FRAME+ 4*8-8(sp)
        std   5,SZ_FRAME+ 5*8-8(sp)
        std   6,SZ_FRAME+ 6*8-8(sp)
        std   7,SZ_FRAME+ 7*8-8(sp)
        std   8,SZ_FRAME+ 8*8-8(sp)
        std   9,SZ_FRAME+ 9*8-8(sp)
        std  10,SZ_FRAME+10*8-8(sp)
        std  11,SZ_FRAME+11*8-8(sp)
        std  12,SZ_FRAME+12*8-8(sp)
        std  13,SZ_FRAME+13*8-8(sp)
        std  14,SZ_FRAME+14*8-8(sp)
        std  15,SZ_FRAME+15*8-8(sp)
        std  16,SZ_FRAME+16*8-8(sp)
        std  17,SZ_FRAME+17*8-8(sp)
        std  18,SZ_FRAME+18*8-8(sp)
        std  19,SZ_FRAME+19*8-8(sp)
        std  20,SZ_FRAME+20*8-8(sp)
        std  21,SZ_FRAME+21*8-8(sp)
        std  22,SZ_FRAME+22*8-8(sp)
        std  23,SZ_FRAME+23*8-8(sp)
        std  24,SZ_FRAME+24*8-8(sp)
        std  25,SZ_FRAME+25*8-8(sp)
        std  26,SZ_FRAME+26*8-8(sp)
        std  27,SZ_FRAME+27*8-8(sp)
        std  28,SZ_FRAME+28*8-8(sp)
        std  29,SZ_FRAME+29*8-8(sp)
        std  30,SZ_FRAME+30*8-8(sp)
        std  31,SZ_FRAME+31*8-8(sp)
        mflr r31  // &decompress
        call unfold
        .long O_BINFO  // .int4
        /* { b_info={sz_unc, sz_cpr, {4 char}}, folded_loader...} */

/* vim:set ts=8 sw=8 et: */
